/* -> c.ckamai
 */

char *versio = "Arthur-Kermit v1.3 (from C-Kermit 4C(052)) 12 Nov 87";

/*  C K A M A I  --  Arthur-C-Kermit Main program  */

/*
 Authors: Frank da Cruz, Bill Catchings, Jeff Damens;
 Columbia University Center for Computing Activities, 1984-85.
 Copyright (C) 1985, Trustees of Columbia University in the City of New York.
 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided
 this copyright notice is retained. 
*/
/*
 The Kermit file transfer protocol was developed at Columbia University.
 It is named after Kermit the Frog, star of the television series THE
 MUPPET SHOW; the name is used by permission of Henson Associates, Inc.
 "Kermit" is also Celtic for "free".
*/

#include "ckcker.h"
#include "ckcdeb.h"
#ifdef ANSI
#include "h.ckatio"
#include "h.ckamis"
#include "h.ckuusr"
#include <string.h>
#endif

/* Text message definitions.. each should be 256 chars long, or less. */

char *hlptxt = "Kermit Server Commands:\n\n\
GET  filespec REMOTE CWD [directory]  REMOTE SPACE [dir]\n\
SEND filespec REMOTE DIR [filespec]   REMOTE HOST command\n\
FINISH        REMOTE DELETE filespec  REMOTE WHO [user]\n\
REMOTE HELP   REMOTE TYPE filespec    BYE\n\n";

char *srvtxt = "\n\
Arthur-C-Kermit server starting.\n\n\
Return to your local machine by typing its escape sequence;\n\
issue further commands from there;\n\
to shut down the Arthur-Kermit server, use command BYE\n\
or FINISH then re-CONNECT.\n\n";

/* Declarations for Send-Init Parameters */

int spsiz = DSPSIZ,                     /* Biggest packet size we can send */
    rpsiz = DRPSIZ,                     /* Biggest we want to receive */
    timint = DMYTIM,                    /* Timeout interval I use */
    rtimo = URTIME,                     /* Timeout I want you to use */
    timef = 0,                          /* Flag to override what you ask */
    npad = MYPADN,                      /* How much padding to send */
    mypadn = MYPADN,                    /* How much padding to ask for */
    chklen = 1,                         /* Length of block check */
    bctr = 1,                           /* Block check type requested */
    bctu = 1,                           /* Block check type used */
    ebq =  MYEBQ,                       /* 8th bit prefix */
    ebqflg = 0,                         /* 8th-bit quoting flag */
    rpt = 0,                            /* Repeat count */
    rptq = MYRPTQ,                      /* Repeat prefix */
    rptflg = 0,                         /* Repeat processing flag */
    capas = 0;                          /* Capabilities */

char padch = MYPADC,                    /* Padding character to send */
    mypadc = MYPADC,                    /* Padding character to ask for */
    seol = MYEOL,                       /* End-Of-Line character to send */
    eol = MYEOL,                        /* End-Of-Line character to look for */
    ctlq = CTLQ,                        /* Control prefix in incoming data */
    myctlq = CTLQ;                      /* Outbound control character prefix */


/* Packet-related variables */

int pktnum = 0,                         /* Current packet number */
    prvpkt = -1,                        /* Previous packet number */
    sndtyp,                             /* Type of packet just sent */
    size,                               /* Current size of output pkt data */
    osize,                              /* Previous output packet data size */
    maxsize,                            /* Max size for building data field */
    spktl;                              /* Length packet being sent */

char sndpkt[MAXPACK*2],                 /* Entire packet being sent */
    recpkt[RBUFL],                      /* Packet most recently received */
    data[MAXPACK+4],                    /* Packet data buffer */
    srvcmd[MAXPACK*2],                  /* Where to decode server command */
    *srvptr,                            /* Pointer to above */
    mystch = SOH,                       /* Outbound packet-start character */
    stchr = SOH;                        /* Incoming packet-start character */


/* File-related variables */

char filnam[50];                        /* Name of current file. */

int nfils;                              /* Number of files in file group */
long fsize;                             /* Size of current file */

/* Communication line variables */

char ttname[50];                        /* Name of communication line. */

int parity,                             /* Parity specified, 0,'e','o',etc */
    flow,                               /* Flow control, 1 = xon/xoff */
    speed = -1,                         /* Line speed */
    turn = 0,                           /* Line turnaround handshake flag */
    turnch = XON,                       /* Line turnaround character */
    duplex = 0,                         /* Duplex, full by default */
    escape = 0241,                      /* Escape character for connect
                                           defaults is F1 */
    delay = DDELAY,                     /* Initial delay before sending */
    mdmtyp = 0;                         /* Modem type (initially none)  */


/* Statistics variables */

long filcnt,                    /* Number of files in transaction */
    flci,                       /* Characters from line, current file */
    flco,                       /* Chars to line, current file  */
    tlci,                       /* Chars from line in transaction */
    tlco,                       /* Chars to line in transaction */
    ffc,                        /* Chars to/from current file */
    tfc;                        /* Chars to/from files in transaction */

int tsecs;                      /* seconds for transaction */

/* Flags */

int deblog = 0,                         /* Flag for debug logging */
    pktlog = 0,                         /* Flag for packet logging */
    seslog = 0,                         /* Session logging */
    tralog = 0,                         /* Transaction logging */
    displa = 0,                         /* File transfer display on/off */
    stdouf = 0,                         /* Flag for output to stdout */
    xflg   = 0,                         /* Flag for X instead of F packet */
    hcflg  = 0,                         /* Doing Host command */
    fncnv  = 1,                         /* Flag for file name conversion */
    binary = 0,                         /* Flag for binary file */
    warn   = 0,                         /* Flag for file warning */
    quiet  = 0,                         /* Be quiet during file transfer */
    local  = 0,                         /* Flag for external tty vs stdout */
    server = 0,                         /* Flag for being a server */
    cnflg  = 0,                         /* Connect after transaction */
    cxseen = 0,                         /* Flag for cancelling a file */
    czseen = 0;                         /* Flag for cancelling file group */

/* Variables passed from command parser to protocol module */

#ifdef ANSI
extern int conect( void );
extern int proto( void );
#else
char parser();                          /* The parser itself */
#endif
char sstate  = 0;                       /* Starting state for automaton */
char *cmarg  = "";                      /* Pointer to command data */
char *cmarg2 = "";                      /* Pointer to 2nd command data */
char **cmlist;                          /* Pointer to file list in argv */

/* Miscellaneous */

char **xargv;                           /* Global copies of argv */
int  xargc;                             /* and argc  */

extern char *dftty;                     /* Default tty name from ckx???.c */
extern int dfloc;                       /* Default location: remote/local */
extern int dfprty;                      /* Default parity */
extern int dfflow;                      /* Default flow control */


/*  M A I N  --  C-Kermit main program  */

#ifdef ANSI
int
#endif
main(argc,argv) int argc; char **argv; {

#ifndef ANSI
    char *strcpy();
#endif

/* Do some initialization */

    xargc = argc;                       /* Make global copies of argc */
    xargv = argv;                       /* ...and argv. */
    sstate = 0;                         /* No default start state. */
    strcpy(ttname,dftty);               /* Set up default tty name. */
    local = dfloc;                      /* And whether it's local or remote. */
    parity = dfprty;                    /* Set initial parity, */
    flow = dfflow;                      /* and flow control. */
    speed = dfbaudr();                  /* Ask BBC MOS for baud rate */
    
/* Look for a UNIX-style command line... */

    if (argc > 1) {                     /* Command line arguments? */
    concb(0);           /* Tim's bit */
        sstate = cmdlin();              /* Yes, parse. */
        if (sstate) {
            proto();                    /* Take any requested action, then */
            if (!quiet) conoll("");     /* put cursor back at left margin, */
            if (cnflg) conect();        /* connect if requested, */
            doexit(GOOD_EXIT);          /* and then exit with status 0. */
        }
    }   
    
/* If no action requested on command line, enter interactive parser */

    cmdini();                           /* Initialize command parser */
#ifdef ANSI
    while( (sstate = parser()) != 0 ) {          /* Loop getting commands. */
#else
    while(sstate = parser()) {          /* Loop getting commands. */
#endif
        if (sstate) proto();            /* Enter protocol if requested. */
    }


#ifdef ANSI
    return 0;
#endif

}
