# CKIKER.MAK, Version 5A(189), Fri Aug  6 17:00:58 1993
#
# This is a makefile for C Kermit using SAS C.  The lint
# line is for Gimpel FlexeLint, a commercial product.
#
# -wo here makes pointer-int conversions a warning instead of an error,
# and is needed if you use prototypes for the Kermit functions.  In
# particular, the last argument of debug() is specified as an int, but
# actual calls often put a pointer there.  (Is there an ANSI-approved
# way to find the numerical equivalent of a pointer value?)
#
# The optimization is set to so, which turns on full optimization.
# As of 5A(157), the data had finally become too big for the small
# data model, so we are forced to large, but only if DYNAMIC is
# not defined.  The DYNAMIC #define causes C Kermit's send and
# receive values to be allocated at run time, and is recommended.
# If DYNAMIC is used, small model can also be used.

CC = sc
.c.o:
	$(CC) $(CFLAGS) $<

# If NOHELP is not defined, you probably need DATA=F, because the amount
# of string data in ckuus2.c gets too much for STRINGMERGE.
# Do NOT define PARSENSE, as the required code is too ugly to copy
# from ckutio.c to ckitio.c

CONFIG = DEF=NODEBUG DEF=NOHELP DEF=NOHEBREW DEF=NOCYRIL STRMER

# The following options are NOT meant to be changed!

OPTS = $(CONFIG) DEF=AMIGA DEF=DYNAMIC DEF=CK_ANSILIBS DEF=CK_ANSIC \
       DEF=NODIAL DEF=MYCURSES DEF=SIG_V

CFLAGS = $(OPTS) DEBUG=L # OPT OPTTIME

LFLAGS =

#
#  All the object files needed to compile C Kermit.
#

OBJS = ckcmai.o ckucmd.o ckuusr.o ckuus2.o ckuus3.o ckuus4.o ckuus5.o \
		 ckuus6.o ckuus7.o ckuusx.o ckuusy.o \
		 ckcpro.o ckcfns.o ckcfn2.o ckcfn3.o ckuxla.o ckicon.o \
		 ckitio.o ckifio.o ckistu.o ckiutl.o \
		 ckcmdb.o

#
# The corresponding sources, used mainly to run lint.
#

SRCS = ckcmai.c ckucmd.c ckuusr.c ckuus2.c ckuus3.c ckuus4.c ckuus5.c \
		 ckuus6.c ckuus7.c ckuusx.c ckuusy.c \
		 ckcpro.c ckcfns.c ckcfn2.c ckcfn3.c ckuxla.c ckicon.c \
		 ckitio.c ckifio.c ckistu.c ckiutl.c

#
# These are prototype files.  You can generate them by saying "make ckipro.h"
#

PROS = ckcmai.pro ckucmd.pro ckuusr.pro ckuus2.pro ckuus3.pro ckuus4.pro ckuus5.pro \
		 ckuus6.pro ckuus7.pro ckuusx.pro ckuusy.pro \
		 ckcpro.pro ckcfns.pro ckcfn2.pro ckcfn3.pro ckuxla.pro ckicon.pro \
		 ckitio.pro ckifio.pro ckistu.pro ckiutl.pro


#
# SAS C rule for generating a prototype file from a C source file.
#

.c.pro:
	sc $(CFLAGS) GENPROTO $*.c GPFILE $*.pro


kermit:    $(OBJS)
	sc link $(LFLAGS) to kermit $(OBJS) MATH=S MAP

ckcpro.o: ckcpro.w wart
	wart ckcpro.w ckcpro.c
	sc $(CFLAGS) to ckcpro.o ckcpro.c

wart: ckwart.c
	sc ckwart.c

lint: $(SRCS)
	lint $(OPTS) $(SRCS)

ckipro.h: $(PROS)
	join $(PROS) as ckipro.h


TAGS: $(SRCS)
	-delete TAGS
	ctags $(SRCS)

clean:
	-delete $(OBJS) kermit wermit wart

#
# Special rules for the Amiga-specific files
#
ckifio.o:	ckifio.c
	$(CC) $(CFLAGS) ckifio.c

ckitio.o:	ckitio.c
	$(CC) $(CFLAGS) DEF=_STRICT_ANSI ckitio.c

ckiutl.o:	ckiutl.c
	$(CC) $(CFLAGS) ckiutl.c

ckicon.o:	ckicon.c
	$(CC) $(CFLAGS) ckicon.c

# Special rules for some other files, uncomment as needed

#ckuus2.o:	 ckuus2.c
#	 $(CC) $(CFLAGS) DATA=F NOSTRMER ckuus2.c
